/* FILE: avftovio.cc                 -*-Mode: c++-*-
 *
 * Program to try to convert from Any (known) Vector Field file
 * to a vio vector field file format.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "oc.h"

#include "template.h"
#include "vector.h"
#include "vecfile.h"

/* End includes */

#define AVFTOVIO_VERSION "1.0"

static char *inname;
static Mesh *mesh;

void Usage(void)
{
  fprintf(stderr,"Usage: avftovio infile >outfile\n");
  exit(1);
}

int
Tcl_AppInit(Tcl_Interp *interp)
{
    Oc_AutoBuf ab;
    if (Oc_Init(interp) == TCL_ERROR) {
        return TCL_ERROR;
    }
    if (Tcl_GlobalEval(interp, 
            ab("Oc_InitScript Avftovio " AVFTOVIO_VERSION)) != TCL_OK) {
        fprintf(stderr,"InitScript error: %s\n", interp->result);
        return Tcl_GlobalEval(interp, ab("exit 1"));
    }

    // Fill mesh from input file
    VectorFieldFileInput *vffreader = VectorFieldFileInput::NewReader(inname);
    if(vffreader==NULL) { // Unknown file type or other error
        fprintf(stderr,"ERROR detected processing file %s\n",inname);
        return Tcl_GlobalEval(interp, ab("exit 1"));
    }
    else { // Reader found for inname
        mesh=vffreader->NewMesh();
        delete vffreader;
        if(mesh==NULL) {
            fprintf(stderr,"ERROR detected while reading file %s\n",inname);
            return Tcl_GlobalEval(interp, ab("exit 2"));
        }
    }

    // Write vio file to stdout
    VioFileOutput vfo;
    vfo.WriteMesh(mesh,stdout,NULL);

    // That's all folks!
    fflush(stdout);
    return Tcl_GlobalEval(interp, ab("exit 0"));
}

int Oc_Main(int argc,char **argv)
{
    if(argc!=2) Usage();
    inname=argv[1];

    Tcl_Main(argc, argv, Tcl_AppInit);
    return 0;
}
