/* FILE: mmdisp.cc                 -*-Mode: c++-*-
 *
 * Mainline for C++/Tcl/Tk 2D vector field display program
 *
 */

#include <string.h>
#include <stdio.h>	/* freopen */
#include <stdlib.h>	/* strcmp (?), ... */

#include "mmdisp.h"
#include "mmdinit.h"	/* Mmdisp_Init, TclCommandsPreInit, ... */

#include <tcl.h>
#include <tk.h>		/* Tk_Main */

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

int Tcl_AppInit(Tcl_Interp *interp)
{
    Oc_AutoBuf ab;
    Tcl_DString buf;

    Oc_SetUseTkFlag(1);
    if (Oc_Init(interp) == TCL_ERROR) {
       return TCL_ERROR;
    }

    if (Mmdisp_Init(interp) == TCL_ERROR) {
        Tcl_DStringInit(&buf);
        Tcl_DStringAppend(&buf,ab("Oc_Log Log {"), -1);
        Tcl_DStringAppend(&buf,interp->result, -1);
        Tcl_DStringAppend(&buf,ab("} error"), -1);
        return Tcl_Eval(interp, Tcl_DStringValue(&buf));
    }
    return TCL_OK;
}

void Usage()
{
  panic(Oc_AutoBuf("Usage: mmdisp [--server] [--nofork]"
		"[--version] [--] [filename]"));
  exit(1);
}

void ShowVersion()
{
  panic(Oc_AutoBuf("mmDisp version: %s\nBuilt against Tcl %s"),
	MMDISP_PATCH_LEVEL,TCL_PATCH_LEVEL);
  exit(1);
}

void CheckParams(int argc,char **argv, char* &filename)
{
  filename=NULL;   // Default is no file

  int i,flagsdone=0;
  for(i=1;i<argc;i++) {
    if(flagsdone || argv[i][0]!='-') {
      // Filename
      if(filename!=NULL) Usage();   // Only 1 file allowed!
      filename=argv[i];
    }
    else if(strcmp(argv[i],"--server")==0) {
      // Save value to be given to interpreter later
      Mmd_SaveServerFlag(1);
    }
    else if(strcmp(argv[i],"--version")==0
	    || strcmp(argv[i],"-v")==0) {
      ShowVersion();
    }
    else if(strcmp(argv[i],"--")==0) {
      flagsdone=1;  // No more command line flags.  Anything else on
      /// command line will be interpreted as a filename.
    }
  }
}

int Oc_Main(int argc,char **argv)
{
    char *filename;
    CheckParams(argc,argv,filename);
    TclCommandsPreInit(argv[0],filename);

    // At this point, argv[0] holds the name of this executable, either
    // as a full path, or relative to the current working directory.
    // Relative to that, find mmdisp.tcl in its parent directory, using
    // the Tcl library functions for platform-independent filename
    // manipulation.
    //
    // This may fail if argv[0] doesn't hold the name of the executable
    // file, but instead the name of a link to it.
    //
    // A better solution here would be to call a Tcl C API function
    // which returned the same value as the Tcl command 
    // [info nameofexecutable], but such a beast is not in the interface.
    //
    // Even better than that would be to structure mmDisp as a plain shell
    // program, which takes a script argument on the command line the same
    // as any Tcl/Tk-based shell program.  That would restore the functioning
    // of -display and other options normally processed by Tk_Main.
    //
    int sargc;
    char **sargv;
    Tcl_SplitPath(argv[0], &sargc, &sargv);

    int jargc = sargc+1;
    char **jargv = new char*[jargc]; 
    Tcl_DString buf;
    int i;
    for (i=0; i<jargc-2; i++) {
        jargv[i] = sargv[i];
    }
    Oc_AutoBuf ab1(".."),ab2("mmdisp.tcl");
    jargv[jargc-2] = (char *)ab1;
    jargv[jargc-1] = (char *)ab2;
    Tcl_DStringInit(&buf);
    Tcl_JoinPath(jargc, jargv, &buf);

    int fakeargc = argc+1;
    char **fakeargv = new char*[fakeargc];
    fakeargv[0]=argv[0];
    fakeargv[1]= new char[Tcl_DStringLength(&buf)+1];
    strcpy(fakeargv[1],Tcl_DStringValue(&buf));

    Tcl_DStringFree(&buf);
    Tcl_Free((char *) sargv);
    delete [] jargv;

    for(i=1;i<argc;i++) fakeargv[i+1]=argv[i];

    Tk_Main(fakeargc,fakeargv,Tcl_AppInit);
    
    // Memory cleanup which will never be reached.
    delete [] fakeargv[1];
    delete [] fakeargv;

    return 0;
}

