/* FILE: omfsh.cc                 -*-Mode: c++-*-
 *
 *	A Tcl shell extended by the OOMMF core (Oc) extension
 *
 * NOTICE: Plase see the file ../../LICENSE
 *
 * Last modified on: $Date: 1998/05/31 21:42:02 $
 * Last modified by: $Author: donahue $
 */

/* Header files for system libraries */
#include <string.h>

/* Header files for third-party extensions */
#include <tcl.h>
#include <tk.h>

/* Header files for the OOMMF core extension */
#include "oc.h"

/* End includes */     

/*
 * Do we load Tk? (Default = yes)
 */
static BOOL useTk = 1;

/* 
 * Removes argument "remove_index" from argument list by
 * translating all arguments after it down one spot and
 * decrementing argc.
 */
static void 
remove_arg(int &argc, char **argv, int remove_index)
{ 
    if(remove_index<0 || remove_index>=argc) return;
    for(int j=remove_index+1;j<argc;j++) {
        argv[j-1]=argv[j];
    }
    argc--;
}

/*
 * Sets static BOOL variable useTk from command line
 */
static void 
ExtractOptions(int &argc, char **argv)
{
    int i=1;
    while(i<argc) {
        if(strcmp(argv[i],"--")==0){
            // End flag processing
            break;
        }

        if(argv[i][0]!='-') {
            // Not an option flag. Under Windows, should one
            // use '/' instead of '-' ??? 
            i++;
            continue;
        }

        char *optflag=&(argv[i][1]);
        if(*optflag=='-') optflag++;  // Treat "-" and "--" the same

        // Known options
        //	Should this processing be case sensitive ???
        if(strcmp(optflag,"tk")==0) {
            useTk=1;
            remove_arg(argc,argv,i);
        }
        else if(strcmp(optflag,"notk")==0 || strcmp(optflag,"no_tk")==0) {
            useTk=0;
            remove_arg(argc,argv,i);
        }
        else {
            // Anything else is left on the command line untouched, for
            // possible use by the shell script
            i++;
        }
    }
}

/* 
 * The "application initialization" for OOMMF applications.
 */
int 
Tcl_AppInit(Tcl_Interp *interp)
{
    Oc_SetUseTkFlag(useTk);
    if (Oc_Init(interp) != TCL_OK) {
        return TCL_ERROR;
    }
    return TCL_OK;
}

int Oc_Main(int argc, char **argv)
{
    ExtractOptions(argc,argv);

    // Turn over control to the Tcl/Tk library.
    if (useTk) {
        Tk_Main(argc, argv, Tcl_AppInit);
    } else {
        Tcl_Main(argc, argv, Tcl_AppInit);
    }

    // Tk_Main and Tcl_Main do not return, so control never gets here.
    // Make picky compilers happy.
    return 0;
}

