/* FILE: stopwatch.cc                   -*-Mode: c++-*-
 *
 * StopWatch class
 *
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 1998/06/04 01:35:23 $
 * Last modified by: $Author: donahue $
 */

#include "stopwatch.h"
#include "errhandlers.h"

/* End includes */     

//////////////////////////////////////////////////////////////////////////
// StopWatch timing class
//
const ClassDoc Nb_StopWatch::class_doc("Nb_StopWatch",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","May-1998");


Nb_StopWatch::Nb_StopWatch(BOOL _running)
{
  running=_running;
  // Initialize Oc_TimeVal's
  Oc_Times(cpu_last,wall_last);
  cpu_total=cpu_last; cpu_total.Reset(); // Make compatible
  wall_total=wall_last; wall_total.Reset(); // Ditto
}

void Nb_StopWatch::Start()
{
  if(running) return;
  // Start clock
  running=1;
  Oc_Times(cpu_last,wall_last);
}

void Nb_StopWatch::Stop()
{
  if(!running) return;  // Watch already stopped
  // Stop clock
  Oc_TimeVal cpu_now,wall_now;
  Oc_Times(cpu_now,wall_now);
  cpu_total+=(cpu_now-cpu_last);
  wall_total+=(wall_now-wall_last);
  running=0;
}

void Nb_StopWatch::GetTimes(Oc_TimeVal &cpu,Oc_TimeVal &wall) const
{
  if(!running) {
    cpu=cpu_total; wall=wall_total;
  } else {
    // Include current accumlated times
    Oc_TimeVal cpu_now,wall_now;
    Oc_Times(cpu_now,wall_now);
    cpu=cpu_total+(cpu_now-cpu_last);
    wall=wall_total+(wall_now-wall_last);
  }
}
